/**
 * FileParser.cpp
 * Author: Yuan Wang
 * Read in an XML file and parse it.
 */

#include <string.h>
#include <iostream.h>
#include <stdlib.h>

#include <util/PlatformUtils.hpp>
#include <internal/URLInputSource.hpp>
#include <parsers/SAXParser.hpp>

#include "FileParser.h"
#include "buffervector.h"


RecordStore* DirPool::recordStore =  new RecordStore();
FileStore* DirPool::fileStore   =  new FileStore();

BufferVector *filenamebv=new BufferVector(256);

/**
 * Constructor.
 * @param rootName      record root name
 * @param indexName     index names
 * @param indexCount    # of indices
 * @param doValidation  parse with validation
 * @param doEscapes     
 */
FileParser::FileParser ( char* rootName, char* indexName[], char* indexType, int* fieldLength,
			 int indexCount, bool doValidation )
{
  // Initialize the XML4C system
  try
    {
      XMLPlatformUtils::Initialize();
    }
  catch ( const XMLException& toCatch )
    {
      cerr << "Error during initialization!\n"
	   << toCatch.getMessage() << endl;
    }

  // Initialize parser and handler.
  parser = new SAXParser();
  parser->setDoValidation( doValidation );
  handler = new FileParserHandler();
  parser->setDocumentHandler( handler );
  parser->setErrorHandler( handler );

  // Initialize YingWu's class.
  //cout << "Initialize YingWu's class ..." << endl;
  this->fileStore = DirPool::fileStore;
  this->recordStore = DirPool::recordStore;

  // Initialize WeiShen's class.
  //cout << "Initialize Weishen's class ..." << endl;
  this->sort = new Sort( indexCount, indexType, fieldLength, indexName, 100 );


  // Transfer names of elements that are being indexed.
  handler->setIndexName( rootName, indexName, indexCount, fieldLength );
  handler->setStorageObject( this->fileStore, this->recordStore, this->sort );
}


/**
 * Destructor.
 */
FileParser::~FileParser ()
{
  delete this->fileStore;
  delete this->recordStore;

  delete this->parser;
  delete this->handler;
}

void FileParser::Run() {
	sort->Run();
}

/**
 * Parse an XML file.
 * @param xmlFile   xml file name
 */
void FileParser::parse ( char* xmlFile )
{
  try
    {
      //cout << "Parse file -- " << xmlFile << endl;

      // Get fid and pass it to parser handler.
      //      int fid = this->fileNames->addFileName( xmlFile );
      //      handler->setFileId( fid );

      parser->parse( xmlFile );
    }
  catch ( const XMLException& toCatch )
    {
      cerr << "\nFile not found: '" << xmlFile << "'\n"
	   << "Exception message is: \n"
	   << toCatch.getMessage() << "\n" << endl;
    }
}


int main ( int argc, char* argv[] )
{
  if ( argc != 2 )
    cerr << "Usage: ---" << endl;
  else
    {
      char rootName[20] = "inproceedings";
      char* indexName[2] = { "year", "author" };
      char indexType[2] = { STRINGTYPE, STRINGTYPE };
      int fieldLength[2] = { 4, 32 };

      FileParser* parser = new FileParser( rootName, indexName, indexType, fieldLength, 2, false );

      FileNameList* fileNameList = new FileNameList( argv[1] );

      char* filename;
      char filename_store[256];
      for(int i=0; i<256; ++i) filename_store[i]='\0';
      
      PageStore::initTime();
      while( ( filename = fileNameList->getNext() ) != NULL )
	{
	  //cout << "File name -------- " << filename << endl;
	  strcpy(filename_store, filename);
	  filenamebv->addValue(filename_store);
	  parser->parse( filename );
	}
      PageStore::showTime();
      PageStore::initTime();
	parser->Run();
      PageStore::showTime();	
/*
	    cout<<"**********************  Test 0 ******************************"<<endl;
        DirPool* dirPool=new DirPool();
	    dirPool->execCmd("cd year=\"1987\"\n");

        cout<<endl;
	    cout<<endl;
        cout<<endl;
	    //delete dirPool;

*/
	    cout<<"**********************  Test 1 ******************************"<<endl;
	    char buffer[256];
        DirPool* dirPool1=new DirPool();
	char c;
	do{
		cout<<"Please input Command:";
		cin.get(buffer,256, '\n');
	    if(strlen(buffer)>0) dirPool1->execCmd(buffer);
	    cin.get(c);
	} while(strlen(buffer)!=0);

        cout<<endl;
	    cout<<endl;
        cout<<endl;
	    delete dirPool1;
    }

    delete DirPool::recordStore;
    delete DirPool::fileStore;
}
