#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "FileStore.h"
#include "RecordStore.h"
#include "IDTable.h"

RecordStore::RecordStore(int idBucket=128, int locBucket=8, int levelBucket=3){
     
     record_counter       =0;
     total_size         =0;
     this->idBucket	= idBucket;
     this->locBucket	= locBucket;
     this->levelBucket	= levelBucket;
 
     current_loc 	= Location();
     idToLoc		= new IDLocationList [idBucket];
     locToId		= new Location_content * [levelBucket];
     for (int i=0;i<levelBucket;i++) 
         locToId[i]  	= new Location_content [locBucket]; 

     initTable();

     idTable 		= new IDTable();

}

RecordStore::~RecordStore(){
     printf("total record counter: %d ,total size %d\n",record_counter, total_size);

     clearTable();
     
     for (int i=0;i<levelBucket;i++) 
        if (locToId[i]!=NULL) delete [] (locToId[i]);

     delete [] idToLoc;
     delete [] locToId;

     delete idTable;
}

void RecordStore::createRecord(char * rootName, int fid, int eid){

     if (current_loc.isRoot()==false){
         printf("Warning: last Record not finished\n");
         clearTable();
         current_loc = Location ();
     }    
     
     clearTable(); 

     unsigned short rootID = idTable->insertElement(rootName, 'e');
     
     // insert to ID to Location Table.
     insertInIDLocationList(rootID,&current_loc); 
     
     // insert to Location to ID Table;
     insertInLocationContentTable(&current_loc, rootID, NULL);
     
     levelCounter 	= 0;
     elemtCounter	= 1;

     current_fid	= fid;
     current_eid	= eid;

}

char * RecordStore::finishRecord(int & size){
     
     if (current_loc.isRoot() == false){
         printf("Warning: the Record did not perfect matched when finish\n");
     }
     
     unsigned short storeElemt  = 4 ;    // scale ??
     
     unsigned short IDSize , LocSize; 

     char * IDString ;
     char * LocString;
     
     IDString  = storeIDTable(storeElemt, IDSize);
     
     //printf("finished here\n");
     LocString = storeLocTable(storeElemt, LocSize);
     
// commented by wei shen
//     printf("IDSize %d LocSize %d\n",IDSize, LocSize);
     
     char  * result = new char [3*sizeof(short)*2*sizeof(int)+ IDSize +LocSize];
     memcpy(result, IDString, IDSize);
     memcpy(result+IDSize, LocString, LocSize);
     
     // memcpy IDSIZE, LOCSize, LEVEL;
     memcpy(result+IDSize+LocSize, (char *)(&IDSize), sizeof(IDSize));
     memcpy(result+IDSize+LocSize+ sizeof(IDSize), 
            (char *)(&LocSize), sizeof(LocSize));
     memcpy(result+IDSize+LocSize+2*sizeof(IDSize), 
            (char *)(&storeElemt), sizeof(storeElemt));
     
     // memcpy fid eid
     size = 3*sizeof(short)+ IDSize +LocSize;
     memcpy(result+size, (char *)(&current_fid), sizeof(current_fid));
     memcpy(result+size+sizeof(current_fid), (char*)(&current_eid), sizeof(current_eid));
     
     size += 2*sizeof(current_fid);

     clearTable();

     delete [] IDString;
     delete [] LocString;
     
     record_counter ++;
     total_size += size;

     return result;
}

void RecordStore::createElement(char * elementName){
     
     unsigned short elementID = idTable->insertElement(elementName, 'e');
     //unsigned short elementID =1;
     
     // insert to ID to Location Table.
     insertInIDLocationList(elementID,&current_loc); 

     // insert to Location to ID Table;
     insertInLocationContentTable(&current_loc, elementID, NULL);

     elemtCounter ++;
}

void RecordStore::addElementContent(char * content){
     
     
     int level		= current_loc.getLevel()%levelBucket;
     int loc_sum	= current_loc.getLocationSum()%locBucket;

     Location_content * loc_content;
     loc_content	= &(locToId[level][loc_sum]);
     
     while(loc_content->loc.sameLocation(&current_loc)==false){
	loc_content = (Location_content *)(loc_content->next);
	if (loc_content ==NULL) break; 
     }	
     
     // find the location;
     if (loc_content!=NULL){
         loc_content->content = strdup(content);
     }else {
         //printf("Error: location not find!!!\n");
     }

}

void RecordStore::finishElement(){
   
    current_loc.growInSameLevel();
}

void RecordStore::addAttribute(int pos, char *attr, char *content){
     
     unsigned short attrID 	= idTable->insertElement(attr,'a');

     Location * attrLoc	= current_loc.createAttributeLocation((char)pos);
     // insert to ID to Location Table.
     insertInIDLocationList(attrID,attrLoc); 
     
     // insert to Location to ID Table;
     insertInLocationContentTable(attrLoc, attrID, content);
     
     delete attrLoc;
}

void RecordStore::goToNextLevel(){
     
     int level		= current_loc.getLevel()%levelBucket;
     int loc_sum	= current_loc.getLocationSum()%locBucket;

     Location_content * loc_content;
     loc_content	= &(locToId[level][loc_sum]);

     while(loc_content->loc.sameLocation(&current_loc)==false){
	loc_content = (Location_content *)(loc_content->next);
	if (loc_content ==NULL) break; 
     }	
     
     // find the location;
     if (loc_content!=NULL){
         loc_content->hasChildren 	= true;
     }else {
         //printf("Error: location not find!!!\n");
     }

     current_loc.increaseLevel(); 

     levelCounter ++;
}

void RecordStore::finishThisLevel(){
     
     current_loc.decreaseInSameLevel();

     int level		= current_loc.getLevel()%levelBucket;
     int loc_sum	= current_loc.getLocationSum()%locBucket;

     Location_content * loc_content;
     loc_content	= &(locToId[level][loc_sum]);

     while(loc_content->loc.sameLocation(&current_loc)==false){
	loc_content = (Location_content *)(loc_content->next);
        if (loc_content == NULL) break;

     }
     // find the location;
     if (loc_content!=NULL){
         loc_content->isLast	 	= true;
     }else {
         //printf("Error: location not find!!!\n");
     }
     
     current_loc.decreaseLevel(); 

     levelCounter --;

}

Location *RecordStore::getCurrentLocation(){
    return & current_loc;
}


///////////////////////////////////////////////////////////
//private  function
///////////////////////////////////////////////////////////

void RecordStore::insertInIDLocationList(unsigned short id, Location *loc){
      
      int index = id % idBucket;
      
      if (idToLoc[index].ID == INVALID) {
            idToLoc[index].ID     	= id;
            idToLoc[index].next   	= NULL;
            idToLoc[index].loc_list	= new LocationList();
            idToLoc[index].loc_list->insertLocation(loc);
            
            return;
      }
      
      IDLocationList *idLocationList = &(idToLoc[index]);
      while ((idLocationList->ID != id)&&(idLocationList->next !=NULL)){
          idLocationList = (IDLocationList *)(idLocationList->next);
      }

      if (idLocationList->ID == id )          
           idLocationList->loc_list->insertLocation(loc);
      else {
         idLocationList->next		= new IDLocationList;
         idLocationList			= (IDLocationList *)(idLocationList->next);
         idLocationList->ID		= id;
         idLocationList->next 		= NULL;
         idLocationList->loc_list 	= new LocationList();
         idLocationList->loc_list->insertLocation(loc);
      }
}

void RecordStore::insertInLocationContentTable(Location * loc, 
        unsigned short id, char * content){
     
     int level 		= loc->getLevel()%levelBucket;
     int loc_sum	= loc->getLocationSum()%locBucket;

     if (locToId[level][loc_sum].ID == INVALID) {
         locToId[level][loc_sum].loc.copyLocation(loc);
         locToId[level][loc_sum].ID		= id;
	 locToId[level][loc_sum].isLast		= false;
	 locToId[level][loc_sum].hasChildren    = false;
         if( content !=NULL)
	     locToId[level][loc_sum].content	= strdup(content);
	 else locToId[level][loc_sum].content	= NULL;    
         locToId[level][loc_sum].next		= NULL;
         
	 //printf("dump when insert\n");
	 //locToId[level][loc_sum].loc.dump();

	 if (loc->isRoot()) locToId[level][loc_sum].isLast = true;

         prev=&(locToId[level][loc_sum]);
         return;
     }
     
     Location_content * loc_content 	= new Location_content;
     (loc_content->loc).copyLocation(loc);
     loc_content->ID			= id;
     loc_content->isLast		= false;
     loc_content->hasChildren   	= false;
     if (content !=NULL) 
        loc_content->content 		= strdup(content);
     else loc_content->content 		= NULL; 				
     loc_content->next			= locToId[level][loc_sum].next;
     locToId[level][loc_sum].next	= loc_content;

     prev 				= loc_content;	
}

void RecordStore::initTable(){
     for (int i=0;i<idBucket;i++){
         idToLoc[i].ID 		= INVALID;
	 idToLoc[i].next	= NULL;	
	 idToLoc[i].loc_list    = NULL;
     }

     for (int i=0;i<levelBucket;i++)
         for (int j=0;j<locBucket;j++){

         locToId[i][j].next 	= NULL;
	 locToId[i][j].ID   	= INVALID;
	 locToId[i][j].loc	= Location();
	 locToId[i][j].content  = NULL;
     }  	 
}

void RecordStore::clearTable(){
    
    for (int i=0;i<idBucket;i++){
    	idToLoc[i].ID = INVALID;
	if (idToLoc[i].loc_list!=NULL){ 
	   delete idToLoc[i].loc_list;
	   idToLoc[i].loc_list = NULL;
        }
	if (idToLoc[i].next!=NULL){
	    
	    IDLocationList * temp1 	=(IDLocationList *)(idToLoc[i].next);
	    IDLocationList * temp2;
	    idToLoc[i].next 		= NULL;
	    while(temp1!=NULL){
	         
		 temp2 = (IDLocationList *)(temp1->next);
		 delete temp1->loc_list;
		 delete temp1;
		 temp1 = temp2;
	    }
	}
	
    }

    for (int i=0;i<levelBucket;i++)
       for (int j=0;j<locBucket;j++){
          
	  locToId[i][j].ID 		= INVALID;
	  if (locToId[i][j].content!=NULL){
	     free(locToId[i][j].content);
	     locToId[i][j].content = NULL;
	  }   
	  if(locToId[i][j].next!=NULL){
	  	Location_content * temp1= (Location_content *)(locToId[i][j].next);
	  	Location_content * temp2;
		locToId[i][j].next	= NULL;	
	  	while(temp1!=NULL){
	      		temp2		= (Location_content *)(temp1->next);
	      		free(temp1->content);
			delete temp1;
	      		temp1 		= temp2;
	        } 
	  }else {
              
	  }
	  
    }   
}

char * RecordStore::storeIDTable(unsigned short level, unsigned short & size){
    
    unsigned short    	sizes[level];
    char   		string[level][1024];

    for (int i=0;i<level;i++) sizes[i]=0;
    
    int index;
    for (int i=0;i<idBucket;i++){
        
	index = i% level;

	IDLocationList * temp = &(idToLoc[i]);
        if ( temp -> ID ==INVALID) continue;
	
	while(temp!=NULL){
            
	    string[index][sizes[index]]='<';
	    memcpy(&(string[index][sizes[index]+1]),
	           (char *)(&(temp->ID)),
		   sizeof(unsigned short));
            sizes[index] += (1+ sizeof(unsigned short));
            
	    temp->loc_list->initCurrent();
	    Location * loc;
	    int        tempSize;
	    while((loc=temp->loc_list->getNext())!=NULL){
		
		string[index][sizes[index]]='@';
		
		char * temp;

		temp = loc->toStoreString(tempSize);
		memcpy(&(string[index][sizes[index]+1]), temp, tempSize);
                delete [] temp;
                
		//printf("tempSize %d\n",tempSize);
		
		sizes[index] += tempSize+1;
	    }
	    string[index][sizes[index]] ='>';
	    sizes[index]++;
	
	    
	    // get next Location_record;
	    temp = (IDLocationList *)(temp ->next);
	}

	//printf("end of for %d and  %d \n",i,index);
    }


    size =0;
    for(int i=0;i<level;i++){
        size += sizes [i];
    }

    char * result = new char [size+sizeof(unsigned short)*level];
    
    int temp =0;
    for (int i=0;i<level;i++){
        memcpy(result+temp, string[i], sizes[i]);
	
	temp+= sizes[i];
	memcpy((char *)(result+size+i*sizeof(unsigned short)),
	       (char *)(&(sizes[i])),
	       sizeof(unsigned short));
    }

    size += sizeof(unsigned short)*level;

    return result;
}


char * RecordStore:: storeLocTable(unsigned short elemt, 
				   unsigned short & size){
     int index;
     
     char  string[elemt][1024];
     unsigned short   sizes[elemt];

     for (int i=0;i<elemt;i++) sizes[i]=0;
     
     for (int i=0;i<levelBucket;i++)
        for (int j=0;j<locBucket;j++){
        
	   index =j % elemt;
            
	   Location_content * temp = &(locToId[i][j]);
	   if (temp->ID==INVALID) continue;
	   
	   //printf("enter :%d %d\n",i,j);

	   while(temp!=NULL){
	              
               string[index][sizes[index]]='<';
	       
	       char * tempString;
	       int    tempSize;
               
	       tempString = (temp->loc).toStoreString(tempSize);

	       //printf("tempsize %d\n",tempSize);
               //if (temp->next==NULL) printf("null 1\n");
	       
	       if (tempString!=NULL)
	          memcpy(&(string[index][sizes[index]+1]),
	              tempString,
		      tempSize);
	       sizes[index] += (1+tempSize);	      
	       
	       delete [] tempString;

	       string[index][sizes[index]] = '&';
	       memcpy(&(string[index][sizes[index]+1]), 
	              (char *)(&(temp->ID)), sizeof(unsigned short));
	
	       sizes[index] +=1+sizeof(unsigned short);

	       if (temp->content!=NULL){
	          strcpy(&(string[index][sizes[index]]),
	              temp->content);
	          sizes[index] += (1+strlen(temp->content));
	       }

	       int isLast, hasChildren;
	       if (temp->isLast) isLast =1; else isLast =0;
	       if (temp->hasChildren) hasChildren =1;
	       else hasChildren =0;
	       string[index][sizes[index]]=(char)(16*isLast+hasChildren);
	       
	       string[index][sizes[index]+1] = '>';
	       
	       sizes[index]+= 2;
	       //if (temp->next==NULL) printf("null 2");
	       temp =(Location_content *)((temp->next));
	       
	       //printf("next record\n");
	   }

	   //printf("out %d %d\n",i,j);

     }

    size =0;
    for(int i=0;i<elemt;i++){
        size += sizes [i];
    }

    char * result = new char [size+sizeof(unsigned short)*elemt];
    
    int temp =0;
    for (int i=0;i<elemt;i++){
        memcpy(result+temp, string[i], sizes[i]);
	
	temp+= sizes[i];
	memcpy((char *)(result+size+i*sizeof(unsigned short)),
	       (char *)(&(sizes[i])),
	       sizeof(unsigned short));
    }

    size += sizeof(unsigned short)*elemt;

    return result;
}

void RecordStore::dump(){
}


/*int main(){
   int size;
   Location *loc;
   
   char * eName;
   char * content;
   
   RecordStore recordStore = RecordStore();
   FileStore   fileStore   = FileStore();
   
   for (int j=0;j<3;j++){
      eName  = strdup("books");
      content = strdup("hello");
      recordStore.createRecord(eName,0,0);
      recordStore.addElementContent(content);
      delete [] eName;
      delete [] content;
      
      
      recordStore.goToNextLevel();
//loc = recordStore.getCurrentLocation();
//loc->dump();
      
      eName = strdup("author");
      recordStore.createElement(eName);
      delete [] eName;
//loc = recordStore.getCurrentLocation();
//loc->dump();
      
      
//loc = recordStore.getCurrentLocation();
//loc->dump();
      
      recordStore.finishElement();
//loc = recordStore.getCurrentLocation();
//loc->dump();
      
      eName = strdup("author");
      recordStore.createElement(eName);
      delete [] eName;
      
      eName = strdup("gender");
      content = strdup("female");
      recordStore.addAttribute(1,eName,content);
      delete [] eName;
      delete[] content;
      
      recordStore.finishElement();
      
      eName = strdup("author");
      content = strdup("naughton");
      recordStore.createElement(eName);
      delete [] eName;

      recordStore.addElementContent(content);
      delete [] content;
      
      recordStore.finishElement();
      
      eName = strdup("date");
      recordStore.createElement(eName);
      delete [] eName;
      
      content = strdup("04/01/2000");
      recordStore.addElementContent(content);
      delete [] content;
      
      recordStore.finishElement();
      
      eName = strdup("date");
      recordStore.createElement(eName);
      delete [] content;
      
      recordStore.finishElement();
      
      
      recordStore.finishThisLevel();
//loc = recordStore.getCurrentLocation();
//loc->dump();   
      content = strdup("dewitt");
      recordStore.addElementContent(content);
      delete[] content;

      char * st =recordStore.finishRecord(size);

      printf("size = %d\n",size);

      for (int i=0;i<size;i++) {
         printf("%d ",(int)st[i]);
         if ((i+1)%15==0) printf("\n");
      }
      printf("\n");

      char * path[2] ={"books","author"};
      char ** result;
      int count;

//printf("pre test: %s %s\n",path[0],path[1]);
      recordStore.loadValue(st,size, path,1 ,result,count);
      printf(" find %d result\n",count);
      for (int i=0;i<count;i++) printf("%s\n",result[i]);
      
      int fid;
      unsigned short rid;
      fileStore.insertRecord(st, (unsigned short)size, fid, rid);
printf("insert in : %d %d\n",fid,rid);      
   }

   return 0;
} */
