#ifndef _RECORDSTORE_H_
#define _RECORDSTORE_H_

#include "LocationList.h"
#include "IDTable.h"

typedef struct {
     Location 		loc;
     unsigned short 	ID;
     char *            	content;
     bool		isLast;
     bool		hasChildren;
     void *            	next;
}Location_content;

typedef struct {
     unsigned short	ID;
     LocationList * 	loc_list;
     void *		next;
}IDLocationList;     

class RecordStore {
   private:
     Location 	        current_loc;
     IDLocationList   * idToLoc;
     Location_content **locToId;
     
     IDTable           * idTable;

     int	        idBucket;
     int                locBucket;
     int                levelBucket;

     int		levelCounter;
     int		elemtCounter;
     
     int 		current_fid,current_eid;

     Location_content * prev;
     // static info
     int		record_counter;
     int		total_size;

   private:
     void clearTable();
     void initTable();

     void insertInIDLocationList(unsigned short, Location *);
     void insertInLocationContentTable(Location *, unsigned short, char *);
     
     char * storeIDTable(unsigned short , unsigned short &);
     char * storeLocTable(unsigned short,  unsigned short&) ;

     LocationList *getLocationList(LocationList *locationList, 
           int current_level, char * record, unsigned short level, 
	   unsigned short rec_size, unsigned short id);

     char * getString(char * record, Location * loc, 
                             unsigned short level, int rec_size);
     bool checkValid(Location *loc, LocationList *locationList);			     
     
   public:
       RecordStore(int idBucket=128, int locBucket=8,int levelBucket=3);
      ~RecordStore();

      void 		createRecord(char * rootName, int fid, int eid); //
      char *	 	finishRecord(int &);

      void 		createElement(char *elemtName); //
      void		addElementContent(char * content); //
      void		finishElement(); //
      void		addAttribute(int pos, char * attrName,char * content);//

      void		goToNextLevel();  //
      void		finishThisLevel(); //

      Location	* 	getCurrentLocation(); //

      void              loadValue(char *, int, char * path[], int level,
                                  char **& result, int &);
     
      int		getFid(char * rec, int rec_size);
      int		getEid(char * rec, int rec_size);
      
      bool 		scanCheck(char *  record, int rec_size,
                                  char *  eName,  char * eContent);
      void 		dump();
      
};   


#endif
