#include <iostream.h>
#include "searchbtree.h"
#include "dirlayer.h"
#include "buffervector.h"
#include "dirpool.h"

extern BufferVector* filenamebv;

bool scanCheck(RecordStore * recordStore, FileStore * fileStore,
          int fid, unsigned short rid, char *eName, char * eContent);
int beginScan(RecordStore *recordStore, FileStore * fileStore,
              char * eName, char * eContent, int& fid, unsigned short& rid, 
	      int& yuanFid, unsigned short& yuanRid );
int getNext(RecordStore * recordStore, FileStore * fileStore,
            char * eName,  char *eContent,int& fid, unsigned short& rid, 
	    int& yuanFid, unsigned short& yuanRid);


DirLayer::DirLayer(DirLayer* father, Predit* p, char* buf) {
	this->father=father;
	if(father!=NULL) {
		father->setSon(this);
	}
	son=NULL;
	pp=p;
	result=NULL;
	length=-1;// indicate that it is the first time
	if(buf!=NULL) {
		buffer=new char[strlen(buf)+1];
		strcpy(buffer, buf);
	}
	else buffer=NULL;
} 

DirLayer::~DirLayer() {
	if(length>0) {
		delete [] result;
	}
	Predit* t;
	while(pp!=NULL) {
		t=pp;
		pp=pp->next;
		delete t;
	}

	if(buffer!=NULL) delete [] buffer;
	
}

void DirLayer::setPredit(Predit* p) {
	pp=p;
}

void DirLayer::setFather(DirLayer* f) {
	father=f;
}

void DirLayer::getResult(char*& rr, int& len) {
	rr=result;
	len=length;
}

void DirLayer::getFatherResult(char*& fatherResult, int& fatherLen) {
	if(father==NULL) {
		fatherResult=NULL;
		fatherLen=-1;
	}
	else {
		father->getResult(fatherResult, fatherLen);
	}
}


void DirLayer::execCmd() {
	char* fatherResult;
	int fatherLen;
	char nextType;
	Predit* p;

	getFatherResult(fatherResult, fatherLen);
	if(fatherLen==0) { 
		// the father layer has been processed, and the result is empty
		// so the result should be empty for this layer
		length=0;
		return;
	}
	p=pp;
	nextType='\0';  // set it to OR
	while(p!=NULL) {
		getPreditResult(p, nextType, fatherResult, fatherLen);
		nextType=p->nextType;
		p=p->next;
	}
//	if(WEISHEN_DEBUG) dumpResult();
}

void DirLayer::showResult() {
	dumpResult();
}

/*

void DirLayer::getInitResult(Predit* p, char*& result, int& length) {
	char tag;

	if(father==NULL) {
		SearchBTree* btree=new SearchBTree(p->field);
		if(p->valueType==INTTYPE) {
			btree->search(&p->iValue, result, length, tag);
		}
		else {
			btree->search(p->sValue, result, length, tag);
		}
		if(tag==0) {
			// no such btree, must scan the record!!!
			BufferVector* bv=new BufferVector(2*(sizeof(int)+sizeof(ushort)));
		}
	}
	else {
		int sizeofInt=sizeof(int);
		int sizeofShort=sizeof(ushort);
		int unitSize=2*(sizeofInt+sizeofShort);
	}
}

*/

void DirLayer::dumpResult() {
	int i,j;
	int fid, wfid;
	int oldFid;
	ushort eid, weid;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);
	int unitSize=2*(sizeofInt+sizeofShort);
	int offset=0;
	char* temp;

	cout<<" The search result: "<<endl;
	oldFid=-1;
	for(i=0; i<length; ++i) {
		memcpy(&fid, result+offset, sizeofInt);
		offset+=unitSize;
		if(fid!=oldFid) {
			temp=filenamebv->getElement(fid);
			cout<<temp<<endl;
			delete temp;
			oldFid=fid;
		}
	}
}

//-----------------------------------------------------------------
// if fatherResult=NULL, the fatherLen indicate whether 
// it is the first time
//-----------------------------------------------------------------
void DirLayer::getPreditResult(Predit* p, char nextType, 
					char* fatherResult, int fatherLen) {
	char* temp;
	char* temp1;
	int len;
	int len1;
	SearchBTree *btree;
	char tag;
	int scanTemp;
	int fid, wfid;
	ushort eid, weid;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);
	int unitSize=2*(sizeofInt+sizeofShort);
	BufferVector* bv=new BufferVector(unitSize);
	char* buf=new char[unitSize];

	if(nextType=='\0') { // Or
		btree=new SearchBTree(p->field, tag);
		if(tag==0) {// success
			if(p->valueType==INTTYPE) {
				btree->search((char*)(&(p->iValue)), p->op,temp, len);
			}
			else {
				btree->search(p->sValue, p->op, temp, len);
			}
			delete btree;

/*
			if(len==0) {
				bv->deleteBuffer();
				delete bv;
				delete [] buf;
				return;
			}
*/
			if(fatherResult!=NULL) {
				interSect(temp, len, fatherResult, fatherLen, temp1, len1);
				if(len!=0) delete [] temp;
			}
			else {
				temp1=temp;
				len1=len;
			}
		}
		else { // no btree
			delete btree;
			if(fatherResult==NULL) {// this is the first layer
				// scan the file
				// put the result to temp1!!!
				scanTemp=beginScan(DirPool::recordStore, DirPool::fileStore, 
								p->field, p->sValue,wfid, weid, fid, eid);
				while(scanTemp!=-1) {
					if(scanTemp==1) {// this record match the condition
						memcpy(buf, &fid, sizeofInt);
						memcpy(buf+sizeofInt, &eid, sizeofShort);
						memcpy(buf+sizeofInt+sizeofShort, &wfid, sizeofInt);
						memcpy(buf+2*sizeofInt+sizeofShort, &weid, sizeofShort);
						bv->addValue(buf);
					}
					scanTemp=getNext(DirPool::recordStore, DirPool::fileStore, 
								p->field, p->sValue,wfid, weid, fid, eid);
				}
			}
			else {
				// scan the father
				scanTemp=sizeofInt+sizeofShort;
				while(scanTemp<fatherLen) {
					memcpy(&wfid, fatherResult+scanTemp, sizeofInt);
					memcpy(&weid, fatherResult+scanTemp+sizeofInt, sizeofShort);
					if(scanCheck(DirPool::recordStore, DirPool::fileStore, 
									wfid, weid, p->field,p->sValue)) {
						bv->addValue(fatherResult+scanTemp-
										sizeofInt-sizeofShort);
					}
					scanTemp+=unitSize;
				}
			}
			len1=bv->getSize();
			if(len1==0) {
				temp1=NULL;
			}
			else 
				temp1=bv->getBuffer();
		}
		if(result==NULL) {
			result=temp1;
			length=len1;
		}
		else {
			merge(temp1, len1, result, length, temp, len);
			if(len1!=0) delete [] temp1;
			if(result>0) delete [] result;
			result=temp;
			length=len;
		}
	}
	else {// and
		if(result==NULL&& length==0) {
			// previous item get null result, the result of and will
			// be null again
			delete bv;
			delete [] buf;
			return;
		}
		// note, it is impossible that the result ==null, and the
		// the nextType is and
		btree=new SearchBTree(p->field, tag);
		if(tag==0) {// success
			if(p->valueType==INTTYPE) {
				btree->search((char*)(&(p->iValue)), p->op, temp1, len1);
			}
			else {
				btree->search(p->sValue, p->op, temp1, len1);
			}
			delete btree;
			if(len1==0){
				delete bv;
				delete [] buf;
				delete [] result;
				length=0;
				return;
			}
			interSect(temp1, len1, result, length, temp, len);
			if(len1>0) delete [] temp1;
			if(length>0) delete [] result;
			result=temp;
			length=len;
		}
		else {
			delete btree;
			// scan the result
			scanTemp=sizeofInt+sizeofShort;
			int tempInt=0;
			while(tempInt<length) {
				memcpy(&wfid, result+scanTemp, sizeofInt);
				memcpy(&weid, result+scanTemp+sizeofInt, sizeofShort);
				if(scanCheck(DirPool::recordStore, DirPool::fileStore, wfid, 
								weid, p->field, p->sValue)) {
					bv->addValue(result+scanTemp-sizeofInt-sizeofShort);
				}
				tempInt++;
				scanTemp+=unitSize;
			}
			if(length>0) delete [] result;
			length=bv->getSize();
			if(length==0) {
				result=NULL;
			}
			else 
				result=bv->getBuffer();
		}
	}
	delete bv;
	delete [] buf;
}


//-----------------------------------------------------------------
// get the union of two set
//-----------------------------------------------------------------

void DirLayer::merge(char* buf1, int len1, char* buf2, int len2, 
						char*& result, int& length) {
	int kk1=0, kk2=0;
	int temp1, temp2;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);
	int unitSize=2*(sizeofInt+sizeofShort);
	BufferVector* bv=new BufferVector(unitSize);
	ushort temp3, temp4;
	length=0;
	
	if(len1==0||len2==0) {
		result=NULL;
		length=0;
	}
	memcpy(&temp1, buf1, sizeofInt);
	memcpy(&temp2, buf2, sizeofInt);
	while(kk1<len1 && kk2<len2) {
		if(temp1<temp2) {
			bv->addValue(buf1);
			kk1++;
			buf1+=unitSize;
			if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
		}
		else if(temp1==temp2) {
			memcpy(&temp3, buf1+sizeofInt, sizeofShort);
			memcpy(&temp4, buf2+sizeofInt, sizeofShort);
			if(temp3<temp4) {
				bv->addValue(buf1);
				kk1++;
				buf1+=unitSize;
				if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
			}
			else if(temp3==temp4) {
				bv->addValue(buf1);
				kk1++;
				kk2++;
				buf1+=unitSize;
				buf2+=unitSize;
				if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
				if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
			}
			else {
				bv->addValue(buf2);
				kk2++;
				buf2+=unitSize;
				if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
			}
		}
		else {
			bv->addValue(buf2);
			kk2++;
			buf2+=unitSize;
			if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
		}
	}
	if(kk1!=len1) bv->addValues(buf1, len1-kk1);
	if(kk2!=len2) bv->addValues(buf2, len2-kk2);
	length=bv->getSize();
	if(length!=0)
		result=bv->getBuffer();
	else {
		result=NULL;
	}
	delete bv;
}

//-----------------------------------------------------------------
// get the intersect of two set
//-----------------------------------------------------------------

void DirLayer::interSect(char* buf1, int len1, char* buf2, int len2,
							char*& result, int& length) {
	int kk1=0, kk2=0;
	int temp1, temp2;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);
	int unitSize=2*(sizeofInt+sizeofShort);
	BufferVector* bv=new BufferVector(unitSize);
	ushort temp3, temp4;
	length=0;
	
	if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
	if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
	while(kk1<len1 && kk2<len2) {
		if(temp1<temp2) {
			kk1++;
			buf1+=unitSize;
			if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
		}
		else if(temp1==temp2) {
			memcpy(&temp3, buf1+sizeofInt, sizeofShort);
			memcpy(&temp4, buf2+sizeofInt, sizeofShort);
			if(temp3<temp4) {
				kk1++;
				buf1+=unitSize;
				if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
			}
			else if(temp3==temp4) {
				bv->addValue(buf1);
				kk1++;
				kk2++;
				buf1+=unitSize;
				buf2+=unitSize;
				if(kk1<len1) memcpy(&temp1, buf1, sizeofInt);
				if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
			}
			else {
				kk2++;
				buf2+=unitSize;
				if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
			}
		}
		else {
			kk2++;
			buf2+=unitSize;
			if(kk2<len2) memcpy(&temp2, buf2, sizeofInt);
		}
	}
	length=bv->getSize();
	if(length!=0)
		result=bv->getBuffer();
	else {
		result=NULL;
	}
	delete bv;
}
	
