#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>

#include "constant.h"
#include "node.h"
#include "head.h"
#include "runfile.h"

int Node::file_num_count=0;

/*------------------------------------------------------------------

	Type 1: ENTRYNODE, used for data entry:
		File format:
			nodetype(char), 
			next pointer(int), -----------to next entry node
			otherfield_len(int)
			fid(int), eid(ushort), wfid(int), weid(ushort), 
			fid, eid, wfid, weid.....
		Node format:
			tag==ENTRYNODE
			key_len=0
			otherfield_len=length of a record
			
	Type 2: LEAFNODE, used for Leaf:
		File format:
			nodetype(char), 
			keytype(char)
			keyLen(int)
			otherfield_len(int)
			file_number(int), offset(int), number of record(int), key
			file_number(int), offset(int), number of record(int), key
			file_number(int), offset(int), number of record(int), key
			..........
		Node format:
			tag==LEAFNODE
			key_len= the length of the key
			otherfield_len=length of a record -length of the key
			
	Type 3: OTHERNODE, used for inter node:
		File format:
			nodetype(char), 
			keytype(char)
			keyLen(int)
			otherfield_len(int)
			file_number(int), 
			key, file_number(int), 
			key, file_number(int), 
			key, file_number(int), 
			..........
		Node format:
			tag==OTHERNODE
			key_len= the length of the key
			otherfield_len=length of a record -length of the key 
							(length of int)
			
------------------------------------------------------------------*/

Node::Node(char tag, char key_type, int key_len, int other_field_len, char* prefix) {
	leafTag=tag;
	this->key_len=key_len;
	this->other_field_len=other_field_len;
	this->key_type=key_type;
	father=NULL;

	filename_prefix=new char[strlen(prefix)+1];
	strcpy(filename_prefix, prefix);

	used_space=0;
	if(leafTag==LEAFNODE) {  
		// one tag, one key_type, one key_len, one other_field
		max_space=PAGESIZE-2*sizeof(char)-2*sizeof(int);
	}
	else if(leafTag==OTHERNODE) { 
		// one tag, one key_type, one key_len, one other_field
		max_space=PAGESIZE-2*sizeof(char)-2*sizeof(int);
	}
	else { //one tag, one next pointer, one other_field
		max_space=PAGESIZE-sizeof(char)-2*sizeof(int);
	}
	
	next=0;

	data=new char[max_space];
	Node::file_num_count++;
	file_num=Node::file_num_count;
	first=file_num;
		
}

Node::~Node() {
	delete [] filename_prefix;
	delete [] data;
}

void Node::reset() {
	Node::file_num_count++;
	file_num=Node::file_num_count;
	used_space=0;
	next=0;
}

void Node::setFather(Node* ff) {
	father=ff;
}

//--------------------------------------------------------------------
//  the pointer returned need to be delete by the caller
//--------------------------------------------------------------------

char* Node::getFileName() {
	char* filename;
	filename=new char[256];
	strcpy(filename, filename_prefix);
	sprintf(filename+strlen(filename), "_%d", file_num);
	return filename;
}


Node* Node::getFather() {
	return father;
}

void Node::outToFile() {
	char* filename=getFileName();
	int temp;
	char* buffer=new char[256];
	int sizeofInt=sizeof(int);
	
	RunFile* ff=new RunFile(filename, CREATENEW);
	if(leafTag==LEAFNODE) {
		buffer[0]=leafTag;
		buffer[1]=key_type;
		memcpy(buffer+2, (char*)(&key_len), sizeofInt);
		temp=2+sizeofInt;
	}
	else if(leafTag==OTHERNODE) {
		buffer[0]=leafTag;
		buffer[1]=key_type;
		memcpy(buffer+2, (char*)(&key_len), sizeofInt);
		temp=2+sizeofInt;
	}
	else {
		buffer[0]=leafTag;
		memcpy(buffer+1, (char*)(&next), sizeofInt);
		temp=1+sizeofInt;
		other_field_len=first;
	}
	memcpy(buffer+temp, (char*)(&other_field_len), sizeofInt);
	ff->insertRecord(buffer, temp+sizeofInt);
	ff->insertRecord(data, used_space);
	delete ff;
	delete [] buffer;
	delete [] filename;
//	if(WEISHEN_DEBUG) dump();
}

int Node::getFileNum() {
	return file_num;
}

void Node::setFileNumZero() {
	file_num=0;
}

int Node::getOffset() {
	return used_space;
}

void Node::dump() {
	int temp=0;;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);
	if(leafTag==LEAFNODE) {
		int number, fileNum;
		ushort offset;
		int key;
		cout<<"LEAFNODE:"<<file_num<<endl;
		while(temp<used_space) {
			memcpy(&number, data+temp, sizeofInt);
			temp+=sizeofInt;
			memcpy(&fileNum, data+temp, sizeofInt);
			temp+=sizeofInt;
			memcpy(&offset, data+temp, sizeofShort);
			temp+=sizeofShort;
			cout<<"Number:"<<number<<"  fileNum:"<<fileNum
							<<"  Offset:"<<offset<<"  key:";
			if(key_type==INTTYPE) {
				memcpy(&key, data+temp, sizeofInt);
				cout<<key<<endl;
			}
			else
				cout<<data+temp<<endl;
			temp+=key_len;
		}
	}
	else if(leafTag==ENTRYNODE) {
		int fid, wfid;
		ushort eid, weid;
		cout<<"ENTRYNODE:"<<file_num<<endl;
		while(temp<used_space) {
			memcpy(&fid, data+temp, sizeofInt);
			temp+=sizeofInt;
			memcpy(&eid, data+temp, sizeofShort);
			temp+=sizeofShort;
			memcpy(&wfid, data+temp, sizeofInt);
			temp+=sizeofInt;
			memcpy(&weid, data+temp, sizeofShort);
			temp+=sizeofShort;
			cout<<"oldFid:"<<fid<<"  oldEid:"<<eid<<"  wFid:"<<wfid
								<<" weid:"<<weid<<endl;
		}
	}
	else {
		cout<<"OTHERNODE:"<<file_num<<endl;
		cout<<"key:     pointer:"<<*(int*)data<<endl;
		temp=sizeofInt;
		int pointer;
		while(temp<used_space) {
			if(key_len==sizeofInt) {
				cout<<"key:"<<*(int*)(data+temp)<<"pointer:"
							<<*(int*)(data+temp+sizeofInt)<<endl;
				temp=temp+sizeofInt+sizeofInt;
			}
			else {
				memcpy(&pointer, data+temp+key_len, sizeofInt);
				cout<<"key:"<<(data+temp)<<"pointer:"
							<<pointer<<endl;
				temp=temp+sizeofInt+key_len;
			}
		}
	}
}
			

int Node::addToLeaf(char* data_p, char* key) {
// the node type must be LEAFNODE
	if(used_space>max_space-key_len-other_field_len) {
		outToFile();
		if(father==NULL) {
			father=new Node(OTHERNODE, key_type, key_len, sizeof(int),
							filename_prefix);
			father->addToOther(file_num);
		}
		reset();
		addToLeaf(data_p, key);
		father->addToOther(key, file_num);
		return 1;
	}
	else {

		assert(other_field_len==2*sizeof(int)+sizeof(ushort));
		memcpy(data+used_space, data_p, other_field_len);
		used_space+=other_field_len;
		memcpy(data+used_space, key, key_len);
		used_space+=key_len;

		return 1;
	}
}

int Node::addToOther(char* key, int file_p) {
	if(used_space>max_space-key_len-other_field_len) {
		char* temp_key=new char[key_len];
		int temp_file_p;

		used_space=used_space-key_len-other_field_len;
		memcpy(temp_key, data+used_space, key_len);
		memcpy(&temp_file_p, data+used_space+key_len, sizeof(int));
		outToFile();

		if(father==NULL){
			Node* temp=new Node(OTHERNODE, key_type, key_len, 
						other_field_len,filename_prefix);
			temp->addToOther(file_num);
			setFather(temp);
		}
		reset();
		father->addToOther(temp_key, file_num);

		addToOther(temp_file_p);
		addToOther(key, file_p);
		delete [] temp_key;
		return 1;
	}
	else {
		memcpy(data+used_space, key, key_len);
		used_space+=key_len;
		memcpy(data+used_space, (char*)(&file_p), other_field_len);
		used_space+=other_field_len;
		return 1;
	}
		
}
int Node::addToOther(int file_p) { // called when it is an empty node
	assert(other_field_len==sizeof(int));
	memcpy(data+used_space,(char*)(&file_p), sizeof(int));
	used_space+=other_field_len; 
	return 1;
}

int Node::addToEntry(char* data_p) { // called when entry node
	if(used_space+key_len<max_space) {
		memcpy(data+used_space, data_p, key_len);
		used_space+=key_len;
	}
	else {
		next=Node::file_num_count+1;
		outToFile();
		// this is for the next entry;
		reset();
		memcpy(data+used_space, data_p, key_len);
		used_space+=key_len;
	}
	return used_space-key_len;
}
