#include "searchnode.h"
#include "runfile.h"
#include "constant.h"
#include "head.h"	
#include "buffervector.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>


SearchNode::SearchNode(char* filename, char* prefix) {
	int length;
	int sizeofChar=sizeof(char);
	int sizeofInt=sizeof(int);
	this->prefix=new char[strlen(prefix)+1];
	strcpy(this->prefix, prefix);
	data=new char[PAGESIZE];
	
	length=RunFile::getFileLen(filename);
	RunFile* ff=new RunFile(filename, READONLY);
	ff->getRecord(&nodeType, sizeofChar);
	
	if(nodeType==OTHERNODE) {
		ff->getRecord(&keyType, sizeofChar);
		ff->getRecord((char*)(&keyLen), sizeofInt);
		ff->getRecord((char*)(&otherLen), sizeofInt);
		assert(otherLen==sizeof(int));
		dataLen=length-2*(sizeofChar+sizeofInt);
	}
	else if(nodeType==LEAFNODE) {
		ff->getRecord(&keyType, sizeofChar);
		ff->getRecord((char*)(&keyLen), sizeofInt);
		ff->getRecord((char*)(&otherLen), sizeofInt);
		assert(otherLen==2*sizeofInt+sizeof(ushort));
		dataLen=length-2*(sizeofChar+sizeofInt);
	}
	else { //ENTRYNODE
		ff->getRecord((char*)(&next), sizeofInt);
		ff->getRecord((char*)(&first), sizeofInt);
		otherLen=(sizeof(int)+sizeof(ushort))*2;
		keyLen=otherLen;
		dataLen=length-sizeofChar-sizeofInt-sizeofInt;
	}
	ff->getRecord(data, dataLen);
	delete ff;
}

SearchNode::SearchNode(char* filename, char* prefix, char& tag) {
	int length;
	int sizeofChar=sizeof(char);
	int sizeofInt=sizeof(int);
	this->prefix=new char[strlen(prefix)+1];
	strcpy(this->prefix, prefix);
	data=new char[PAGESIZE];
	
	length=RunFile::getFileLen(filename);
	RunFile* ff=new RunFile(filename, READONLY,tag);
	if(tag==0) return;
	ff->getRecord(&nodeType, sizeofChar);
	
	if(nodeType==OTHERNODE) {
		ff->getRecord(&keyType, sizeofChar);
		ff->getRecord((char*)(&keyLen), sizeofInt);
		ff->getRecord((char*)(&otherLen), sizeofInt);
		assert(otherLen==sizeof(int));
		dataLen=length-2*(sizeofChar+sizeofInt);
	}
	else if(nodeType==LEAFNODE) {
		ff->getRecord(&keyType, sizeofChar);
		ff->getRecord((char*)(&keyLen), sizeofInt);
		ff->getRecord((char*)(&otherLen), sizeofInt);
		assert(otherLen==2*sizeofInt+sizeof(ushort));
		dataLen=length-2*(sizeofChar+sizeofInt);
	}
	else { //ENTRYNODE
		ff->getRecord((char*)(&next), sizeofInt);
		ff->getRecord((char*)(&first), sizeofInt);
		otherLen=(sizeof(int)+sizeof(ushort))*2;
		keyLen=otherLen;
		dataLen=length-sizeofChar-sizeofInt-sizeofInt;
	}
	ff->getRecord(data, dataLen);
	delete ff;
}

SearchNode::~SearchNode() {
	delete [] prefix;
	delete [] data;
}

void SearchNode::reset(char* filename) {
	int length;
	int sizeofChar=sizeof(char);
	int sizeofInt=sizeof(int);
/*
	this->prefix=new char[strlen(prefix)+1];
	strcpy(this->prefix, prefix);
*/
	
	length=RunFile::getFileLen(filename);
	RunFile* ff=new RunFile(filename, READONLY);
	ff->getRecord(&nodeType, sizeofChar);
	
	if(nodeType==OTHERNODE) {
		ff->getRecord(&keyType, sizeofChar);
		ff->getRecord((char*)(&keyLen), sizeofInt);
		ff->getRecord((char*)(&otherLen), sizeofInt);
		dataLen=length-2*(sizeofChar+sizeofInt);
	}
	else if(nodeType==LEAFNODE) {
		ff->getRecord(&keyType, sizeofChar);
		ff->getRecord((char*)(&keyLen), sizeofInt);
		ff->getRecord((char*)(&otherLen), sizeofInt);
		dataLen=length-2*(sizeofChar+sizeofInt);
	}
	else { //ENTRYNODE
		ff->getRecord((char*)(&next), sizeofInt);
		ff->getRecord((char*)(&first), sizeofInt);
		otherLen=(sizeofInt+sizeof(short))*2;
		keyLen=otherLen;
		dataLen=length-sizeofChar-sizeofInt-sizeofInt;
	}
	ff->getRecord(data, dataLen);
	delete ff;
}


void SearchNode::searchBT(char* key, int op, int*& fids, ushort*& eids, 
				int*& wfids, ushort*& weids, int& length) {
	int fileNum;
	ushort offset;
	char *pre, *post;
	char* currentKey;
	int step;
	int temp,temp1;
	char* filename;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);

	switch(nodeType) {
		case LEAFNODE:
			offset=0;
			currentKey=data+2*sizeofInt+sizeofShort;
			step=sizeofInt*2+sizeofShort+keyLen;
			while(offset<dataLen) {
				if((temp=keyCompare(currentKey, key)) <0) {
					currentKey+=step;
				}
				else if(temp==0) {
					// find the entry node
					char* tt;
					tt=data+offset;
					memcpy(&length, tt, sizeofInt);
					tt+=sizeofInt;
					memcpy(&fileNum, tt, sizeofInt);
					tt+=sizeofInt;
					memcpy(&offset, tt, sizeofShort);
					filename=new char[256];
					strcpy(filename, prefix);
					sprintf(filename+strlen(prefix), "_%d", fileNum);
					reset(filename);
					delete [] filename;
					return searchEntry(fileNum, offset, op, fids, eids, wfids, 
									weids, length);
				}
				else if(temp>0) {
					length=0;
					return;
				}
				offset+=step;
			}
			length=0;
			return;
			break;
		case OTHERNODE:
			offset=otherLen;
			pre=data;
			currentKey=pre+otherLen;
			post=currentKey+keyLen;
			step=otherLen+keyLen;
			while(offset<dataLen) {
				offset+=step;
				if((temp=keyCompare(currentKey, key))>0) {
					// the pre pointer
					memcpy(&temp1, pre, sizeofInt);
					filename=new char[256];
					strcpy(filename, prefix);
					sprintf(filename+strlen(prefix), "_%d", temp1);
					reset(filename);
					delete [] filename;
					return searchBT(key, op, fids, eids, wfids, weids, length);
				}
				else if(temp==0) {
					memcpy(&temp1, post, sizeofInt);
					filename=new char[256];
					strcpy(filename, prefix);
					sprintf(filename+strlen(prefix), "_%d", temp1);
					reset(filename);
					delete [] filename;
					return searchBT(key, op, fids, eids, wfids, weids, length);
				}
				else {
					pre=post;
					currentKey+=step;
					post+=step;
				}
			}
			memcpy(&temp1, pre, sizeofInt);
			filename=new char[256];
			strcpy(filename, prefix);
			sprintf(filename+strlen(prefix), "_%d", temp1);
			reset(filename);
			delete [] filename;
			return searchBT(key, op, fids, eids, wfids, weids, length);
			break;
		case ENTRYNODE:
			// error!!!!!!!!!!!!!!!!
			break;
		default:
			break;
	}
}

void SearchNode::searchBT(char* key, int op, char*& buf, int& length) {
	int fileNum;
	ushort offset;
	char *pre, *post;
	char* currentKey;
	int step;
	int temp,temp1;
	char* filename;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);

	switch(nodeType) {
		case LEAFNODE:
			offset=0;
			currentKey=data+2*sizeofInt+sizeofShort;
			step=sizeofInt*2+sizeofShort+keyLen;
			while(offset<dataLen) {
				if((temp=keyCompare(currentKey, key)) <0) {
					currentKey+=step;
				}
				else if(temp==0) {
					// find the entry node
					char* tt;
					tt=data+offset;
					memcpy(&length, tt, sizeofInt);
					tt+=sizeofInt;
					memcpy(&fileNum, tt, sizeofInt);
					tt+=sizeofInt;
					memcpy(&offset, tt, sizeofShort);
					filename=new char[256];
					strcpy(filename, prefix);
					sprintf(filename+strlen(prefix), "_%d", fileNum);
					reset(filename);
					delete [] filename;
					return searchEntry(fileNum, offset, op, buf, length);
				}
				else if(temp>0) {
					length=0;
					return;
				}
				offset+=step;
			}
			length=0;
			return;
			break;
		case OTHERNODE:
			offset=otherLen;
			pre=data;
			currentKey=pre+otherLen;
			post=currentKey+keyLen;
			step=otherLen+keyLen;
			while(offset<dataLen) {
				offset+=step;
				if((temp=keyCompare(currentKey, key))>0) {
					// the pre pointer
					memcpy(&temp1, pre, sizeofInt);
					filename=new char[256];
					strcpy(filename, prefix);
					sprintf(filename+strlen(prefix), "_%d", temp1);
					reset(filename);
					delete [] filename;
					return searchBT(key,op,  buf, length);
				}
				else if(temp==0) {
					memcpy(&temp1, post, sizeofInt);
					filename=new char[256];
					strcpy(filename, prefix);
					sprintf(filename+strlen(prefix), "_%d", temp1);
					reset(filename);
					delete [] filename;
					return searchBT(key, op, buf, length);
				}
				else {
					pre=post;
					currentKey+=step;
					post+=step;
				}
			}
			memcpy(&temp1, pre, sizeofInt);
			filename=new char[256];
			strcpy(filename, prefix);
			sprintf(filename+strlen(prefix), "_%d", temp1);
			reset(filename);
			delete [] filename;
			return searchBT(key, op, buf, length);
			break;
		case ENTRYNODE:
			// error!!!!!!!!!!!!!!!!
			break;
		default:
			break;
	}
}


void SearchNode::searchEntry(int fileNum, ushort offset, int op, int*& fids, ushort*& eids, 
							int*& wfids, ushort*& weids, int& length) {

	int i;
	char* filename;
	int sizeofInt=sizeof(int);
	int sizeofShort=sizeof(ushort);
	char* temp;

	fids=new int[length];
	eids=new ushort[length];
	weids=new ushort[length];
	wfids=new int[length];

	cout<<"error, this should not be called"<<endl;
	temp=data+offset;
	for(i=0; i<length; ++i) {
		if(offset>=dataLen) { 
			// need to go to next file
			filename=new char[256];
			strcpy(filename, prefix);
			sprintf(filename+strlen(prefix), "_%d", next);
			reset(filename);
			delete [] filename;
			offset=0;
			temp=data;
		}
		memcpy(&(fids[i]),temp, sizeofInt);
		temp+=sizeofInt;
		memcpy(&(eids[i]), temp, sizeofShort);
		temp+=sizeofShort;
		memcpy(&(wfids[i]),temp, sizeofInt);
		temp+=sizeofInt;
		memcpy(&(weids[i]), temp, sizeofShort);
		temp+=sizeofShort;
		offset=offset+otherLen;
	}
	return;
}
	
	
void SearchNode::searchEntry(int fileNum, ushort offset, int op, char*& buf, int& length) {

	int i;
	char* filename;
	int record_len=sizeof(int)+sizeof(ushort);
	record_len+=record_len;
	assert(otherLen==record_len);
	BufferVector *bv=new BufferVector(record_len);
	buf=new char[otherLen*length];
	int offset1=0;
	char* temp;
	int currentNum=first;
	filename=new char[256];

	if(op==LE_OP||op==LT_OP) {
		while(currentNum!=fileNum) {
			strcpy(filename, prefix);
			sprintf(filename+strlen(prefix), "_%d", currentNum);
			reset(filename);
			bv->addValues(data, dataLen/record_len);
			assert(dataLen=dataLen/record_len*record_len);
			currentNum=next;
		}
		strcpy(filename, prefix);
		sprintf(filename+strlen(prefix), "_%d", currentNum);
		reset(filename);
		bv->addValues(data, offset/record_len);
	}
		
	if(op!=LT_OP) {
		int tag=0;
		temp=data+offset;
		i=(dataLen-offset)/record_len;
		while(tag==0) {
			if(i<length) {
				if(op==EQ_OP||op==GE_OP||op==LE_OP) bv->addValues(temp, i);
				length-=i;
				strcpy(filename, prefix);
				sprintf(filename+strlen(prefix), "_%d", next);
				reset(filename);
				i=dataLen/record_len;
				temp=data;
			}
			else {
				if(op==EQ_OP||op==GE_OP||op==LE_OP) bv->addValues(temp, length);
				tag=1;
			}
		}	
	}
	if(op==GE_OP||op==GT_OP) {
		if(i>length) {
			temp+=length*record_len;
			bv->addValues(temp, i-length);
		}
		currentNum=next;
		while(currentNum!=0) {
			strcpy(filename, prefix);
			sprintf(filename+strlen(prefix), "_%d", next);
			reset(filename);
			bv->addValues(data, dataLen/record_len);
			assert(dataLen=dataLen/record_len*record_len);
			currentNum=next;
		}
	}

	length=bv->getSize();
	if(length==0) buf=NULL;
	else buf=bv->getBuffer();

	delete bv;
}
	
	


int SearchNode::keyCompare(char* key1, char* key2) {
	if(keyType==INTTYPE) {
		int temp1, temp2;
		memcpy(&temp1, key1, sizeof(int));
		memcpy(&temp2, key2, sizeof(int));
		if(temp1<temp2) return -1;
		else if(temp1==temp2) return 0;
		else return 1;
	}
	else { // STRING TYPE
		return strncmp(key1, key2, keyLen);
	}
}
