
#include "sort.h"
#include "head.h"
#include "constant.h"
#include "runfile.h"
#include "buildbtree.h"
#include "node.h"
#include <iostream.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "dirpool.h"
#include "FileStore.h"
#include <sys/time.h>

#define TIME_DEBUG		1

// these five global variable are used when the function
// tupleCmp is called. The function tupleCmp must use
// these five global variables to compare the two
// tuples.
char gb_sort_order;
int gb_sort_offset;
int gb_sort_size;
char gb_sort_type;
int gb_step_no;

int tupleCmp(const void *tuple1, const void* tuple2);
int tupleCmp1(const void *tuple1, const void* tuple2);


///////////////////////////////////////////////////////////////////
//  This is the constructor of the Sort Class. In this class
///////////////////////////////////////////////////////////////////

Sort::Sort(int num_of_field, char* fieldType, int* fieldSize, 
				char** prefix, int amt_of_buf) {

	char* filename;

	field_type=new char[num_of_field];
	field_size=new int[num_of_field];
	record_len=new int[num_of_field];
	id_len=(sizeof(int)+sizeof(ushort))*2;
	runfile=new RunFile*[num_of_field];
	for(amt_of_field=0; amt_of_field<num_of_field; ++amt_of_field) {
		field_type[amt_of_field]=fieldType[amt_of_field];
		field_size[amt_of_field]=fieldSize[amt_of_field];
		record_len[amt_of_field]=id_len+fieldSize[amt_of_field];
		if(amt_of_field!=0) {
			genFileName(amt_of_field,0,0, filename,1);
			//OutputToFile(sort_area, current_offset, filename);
			runfile[amt_of_field]=new RunFile((char*)filename, 
								(char) CREATENEW);
			delete [] filename;
		}
	}

	btrees=new BuildBTree*[num_of_field];
	for(int i=0; i<num_of_field; ++i) 
		btrees[i]=new BuildBTree(field_type[i], field_size[i], 
							id_len, prefix[i]);
	
	gb_sort_order=ASCENDING;
	this->amt_of_buf=amt_of_buf;
	sort_area=new char[amt_of_buf*PAGESIZE];	
	max_offset=amt_of_buf*PAGESIZE-record_len[0];
	current_offset=0;
	gb_sort_offset=id_len;
	gb_sort_size=field_size[0];
	gb_sort_type=field_type[0];
	gb_step_no=0;

	file_no=0;
	record_count=0;
	map=new Map();
	currentFid=0;
	currentEid=0;
	sizeofInt=sizeof(int);
	sizeofShort=sizeof(ushort);
}

Sort::~Sort() {
	delete [] sort_area;
	delete [] field_type;
	delete [] field_size;
	delete [] record_len;
	delete [] runfile;
	delete map;
	for(int i=0; i<amt_of_field; ++i) delete btrees[i];
	delete [] btrees;
}


void Sort::insertRecord(int field_no, int fid, ushort eid, 
				int wfid, ushort weid, char* value) {
	//field==0, cluster

	char* filename;
	char* buf;
	int i=0;
	

	if(field_no==0) {
		if(fid==currentFid) {
			currentEid=eid;
		}
		else {
			map->setSize(currentEid+1);
			currentEid=eid;
			currentFid=fid;
		}
		if(current_offset<=max_offset) {
			memcpy(sort_area+current_offset, (char*)(&fid), sizeofInt);
			current_offset+=sizeofInt;
			memcpy(sort_area+current_offset, (char*)(&eid), sizeofShort);
			current_offset+=sizeofShort;
			memcpy(sort_area+current_offset, (char*)(&wfid), sizeofInt);
			current_offset+=sizeofInt;
			memcpy(sort_area+current_offset, (char*)(&weid), sizeofShort);
			current_offset+=sizeofShort;
			memcpy(sort_area+current_offset, value, field_size[0]);
			current_offset+=field_size[0];
			record_count++;
		}
		if(current_offset>max_offset) {// the buffer is full, need to write out

			//sort
			qsort(sort_area,record_count, record_len[0], tupleCmp);

			genFileName(0,0,file_no, filename);
			//OutputToFile(sort_area, current_offset, filename);
			RunFile* ff=new RunFile((char*)filename, (char) CREATENEW);
			ff->insertRecord(sort_area, current_offset);
			delete ff;
			delete [] filename;
			file_no++;
			current_offset=0;
			record_count=0;
		}
	}
	else {
		buf=new char[record_len[field_no]];
		memcpy(buf, (char*)(&fid), sizeofInt);
		i=sizeofInt;
		memcpy(buf+i, (char*)(&eid), sizeofShort);
		i+=sizeofShort;
		memcpy(buf+i, (char*)(&wfid), sizeofInt);
		i+=sizeofInt;
		memcpy(buf+i, (char*)(&weid), sizeofShort);
		i+=sizeofShort;
		memcpy(buf+i, value, field_size[field_no]);
		runfile[field_no]->insertRecord(buf, record_len[field_no]);

		delete [] buf;
	}
}

void Sort::Run() {

	int oldFid,i;
	struct timeval tval1, tval2;

	oldFid=DirPool::fileStore->createNewFile();
	map->setSize(currentEid+1);

	if(TIME_DEBUG) {
		RunFile::setTimeZero();
		gettimeofday(&tval1, 0);
	}
	Step0();
	if(TIME_DEBUG) {
		gettimeofday(&tval2, 0);
		cout<<"the First Index"<<endl;
		cout<<"Total time:"<<(tval2.tv_sec-tval1.tv_sec)*1000000+
					tval2.tv_usec-tval1.tv_usec<<endl;
		RunFile::printTime();
	}

	for(i=1; i<amt_of_field; i++) {
		delete runfile[i];
		if(TIME_DEBUG) { 
			RunFile::setTimeZero();
			gettimeofday(&tval1, 0);
		}
		OtherStep(i);
		if(TIME_DEBUG) {
			gettimeofday(&tval2, 0);
			cout<<"The "<<i<<" index"<<endl;
			cout<<"Total time:"<<(tval2.tv_sec-tval1.tv_sec)*1000000+
					tval2.tv_usec-tval1.tv_usec<<endl;
			RunFile::printTime();
		}
	}

	// from 0 to oldFid-1, delete old file
	for(i=0; i<oldFid; ++i) {
		DirPool::fileStore->removeFile(i);
	}
}


void Sort::debug() {
/*
	RunFile* ff=new RunFile("pass30",READONLY);
	int i;
	sortStruct* temp=new sortStruct();
	char* temp1;
	temp1=(char*)temp;

	do {
		if ((i=ff->getRecord(temp1, record_len))!=0)
			cout<<temp->fid<<"  "<<temp->eid<<"  "<<temp->wfid<<"  "<<temp->weid<<"  "<<temp->value<<endl;
		
	}while(i!=0);
*/
}

void Sort::Step0() {

	char* filename;
	char endFlag=0;

	int in_file_no;
	int out_file_no;
	int pass_no;


	if(current_offset!=0) { // if there is still some in the memory

		qsort(sort_area,record_count, record_len[0], tupleCmp);

		if(file_no==0) {
			// only one file, so the sort ended!
			// so we can now adjust the file, read from ying's old file
			// generate new yid, yeid, and write to new file
			char* record;
			char* oldrecord;
			ushort* old_offset;
			int* old_fileNum;
			int* record_count;
			int offset=0;
			int wfid, fid;
			ushort eid, weid;

			record=sort_area;
			oldrecord=record;
			char* temp_data;
			temp_data=new char[sizeofInt+sizeofInt+sizeofShort];
			record_count=(int*)temp_data;
			old_fileNum=(int*)(temp_data+sizeofInt);
			old_offset=(ushort*)(temp_data+sizeofInt+sizeofInt);
			*old_fileNum=0;
			*old_offset=0;
			*record_count=0;
			while(offset<current_offset) {
				memcpy(&fid, record, sizeofInt);
				memcpy(&eid, record+sizeofInt, sizeofShort);
				memcpy(&wfid, record+sizeofInt+sizeofShort, sizeofInt);
				memcpy(&weid, record+2*sizeofInt+sizeofShort, sizeofShort);
				// write new file, get new yfid, yeid---------------+++++
                
 				char * ying_record;
                ushort ying_length; 

                DirPool::fileStore->getRecord(ying_record,ying_length, 	
										wfid, weid);
				DirPool::fileStore->insertRecord(ying_record,ying_length, 
										wfid,weid);
/*
printf("yuan: %d %d",DirPool::recordStore->getFid(ying_record, ying_length),
     				   DirPool::recordStore->getEid(ying_record, ying_length));					
printf("ying: %d %d   Old: %d, %d\n",wfid, weid, fid, eid);
*/
				map->setValue(fid, eid, wfid, weid);

             	delete [] ying_record;
  
				memcpy(record+sizeofInt+sizeofShort, (char*)(&wfid),sizeofInt);
				memcpy(record+2*sizeofInt+sizeofShort, (char*)(&weid), 
											sizeofShort);
				// buildbtree
				if(*old_fileNum==0) {
					*old_fileNum=btrees[0]->getFileNum();
					*old_offset=btrees[0]->addToEntry(record);
					*record_count=1;
					oldrecord=record;
				}
				else if(tupleCmp1(oldrecord, record)!=0) {
					btrees[0]->addToLeaf(temp_data, oldrecord+id_len);
					*old_fileNum=btrees[0]->getFileNum();
					*old_offset=btrees[0]->addToEntry(record);
					*record_count=1;
					oldrecord=record;
				}
				else {
					btrees[0]->addToEntry(record);
					(*record_count)++;
				}
				offset+=record_len[0];
				record=record+record_len[0];
			}
			btrees[0]->addToLeaf(temp_data, oldrecord+id_len);
			btrees[0]->finish();
			delete [] temp_data;
		}
		genFileName(0,0,file_no, filename);
		//!!!!!!!!OutputToFile(sort_area, current_offset, filename);
		RunFile *ff=new RunFile(filename, CREATENEW);
		ff->insertRecord(sort_area, current_offset);
		delete ff;
		delete [] filename;
		file_no++;
		current_offset=0;
		record_count=0;
	}

	if(file_no==1) { // only one file is generated
		return;
	}

	in_file_no=file_no;
	pass_no=1;
	while(!endFlag) {
		Pass(in_file_no, out_file_no, pass_no, endFlag);
		in_file_no=out_file_no;
		pass_no++;
	}
}

void Sort::OtherStep(int step_no){
	int length;
	int count;
	int total_length;

	int in_file_no;
	int out_file_no;
	int pass_no;
	char* filename;
	char endFlag=0;
	
	gb_step_no=step_no;
	gb_sort_size=field_size[step_no];
	gb_sort_type=field_type[step_no];
	in_file_no=0;
	pass_no=0;
	
	genFileName(step_no,0,0, filename,1);
	int filelen=RunFile::getFileLen(filename);
	RunFile *in_file=new RunFile(filename, READONLY);
	total_length=amt_of_buf*PAGESIZE;
	if(total_length>=filelen) {
		endFlag=1;
	}
	else {
		endFlag=0;
	}
	count=total_length/record_len[step_no];
	total_length=count*record_len[step_no];
	while((length=in_file->getRecord(sort_area, total_length))!=0) {
		count=length/record_len[step_no];
		qsort(sort_area,count, record_len[step_no], tupleCmp);

		if(endFlag==1) {
			int offset=0;
			char *record=sort_area;
			char *oldrecord=sort_area;
			char* temp_data;
			int *fileNum;
			ushort *fileOffset;
			int* recordCount;

			temp_data=new char[sizeofInt+sizeofInt+sizeofShort];
			recordCount=(int*)temp_data;
			fileNum=(int*)(temp_data+sizeofInt);
			fileOffset=(ushort*)(temp_data+sizeofInt+sizeofInt);
			*fileNum=0;
			*fileOffset=0;
			*recordCount=0;

			while(offset<length) {
				if(*fileNum==0) {
					*fileNum=btrees[gb_step_no]->getFileNum();
					*fileOffset=btrees[gb_step_no]->addToEntry(record);
					*recordCount=1;
					oldrecord=record;
				}
					
				else if(tupleCmp1(oldrecord, record)!=0) {
					btrees[gb_step_no]->addToLeaf(temp_data, 	
										oldrecord+id_len);
					*fileNum=btrees[gb_step_no]->getFileNum();
					*fileOffset=btrees[gb_step_no]->addToEntry(record);
					*recordCount=1;
					oldrecord=record;
				}
				else {
					btrees[gb_step_no]->addToEntry(record);
					(*recordCount)++;
				}
				offset+=record_len[step_no];
				record+=record_len[step_no];
			}
			btrees[gb_step_no]->addToLeaf(temp_data, oldrecord+id_len);
			btrees[gb_step_no]->finish();
			delete [] temp_data;
		}
		else {
			genFileName(gb_step_no, 0, in_file_no, filename);
			RunFile* ff=new RunFile(filename, CREATENEW);
			ff->insertRecord(sort_area, length);
			delete [] filename;
			delete ff;
		}
		in_file_no++; 
	}
	in_file->deleteFile();
	delete in_file;

	if(in_file_no==1) { // finished sorting
		return;
	}

	pass_no++;
	while(!endFlag) {
		Pass(in_file_no, out_file_no, pass_no, endFlag);
		in_file_no=out_file_no;
		pass_no++;
	}
}

///////////////////////////////////////////////////////////////////
// This method is designed for the later pass(not the first pass,
///////////////////////////////////////////////////////////////////

void Sort::Pass(int in_file_no, int& out_file_no,
				 int pass_no,char& end_flag) {

  int count=0;
  int merge_no;
  RunFile** hh;
  RunFile* out;
  int i;
  char* temp;

  if(in_file_no<amt_of_buf) {
    // the last merge
    end_flag=1;
  }
  else {
    end_flag=0;
  }

  out_file_no=0;
  hh=new RunFile*[amt_of_buf-1];
  // still some in file not merged in the pass
  while(count<in_file_no) {
    // open the output file

    // read the pageno -1 page in to the memory, 
    // ready to merge to output
    merge_no=0;
    while(merge_no<amt_of_buf-1&&count<in_file_no) {
      genFileName(gb_step_no,pass_no-1, count, temp);
      hh[merge_no]=new RunFile(temp, READONLY);
	  delete [] temp;
      merge_no++;
      count++;
    }

    // merge the amt_of_buf files to one file
	genFileName(gb_step_no,pass_no, out_file_no, temp);
	out=new RunFile(temp, CREATENEW);
	delete [] temp;
    one_merge(hh,out,merge_no, end_flag);

    delete out;
    
    for(i=0;i<merge_no;++i) {
      // delete the temperary input file
      hh[i]->deleteFile();

      delete hh[i];
    }
    out_file_no++;
  }

  delete [] hh;
  return ;
}

///////////////////////////////////////////////////////////////////
// This method read from amt_of_buf HeapFiles, and merge them into
// one Heapfile. the input filename is stored in the parameter
// infile, and the output filename is stored in the field outfile
// the number of the input file is stored in the parameter file_no,
// and the last field is the length of the record
///////////////////////////////////////////////////////////////////

void Sort::one_merge(RunFile** infile, RunFile* out,int file_no,
					int end_flag) {
  char** record;
  bool* end_tag;
  int i;
  int tempI;
  char* oldrecord;
  char* tempdata;
  char *temp;
  int wfid, fid;
  ushort eid, weid;

  record=new char*[file_no];
  // allocate the memory to store the record
  for(i=0;i<file_no;++i) {
    record[i]=new char[record_len[gb_step_no]];
	if(record[i]==NULL)
		cout<<"allocate memory error!"<<endl;
  }

  oldrecord=new char[record_len[gb_step_no]];
  tempdata=new char[2*sizeofInt+sizeofShort];
  int* record_count=(int*)(tempdata);
  int* old_fileNum=(int*)(tempdata+sizeofInt);
  ushort* old_offset=(ushort*)(tempdata+sizeofInt+sizeofInt);
  // the filenum can't be 0, so it can be used as a start tag
  *old_fileNum=0; 
  *record_count=0;
  end_tag=new bool[file_no];


  bool* read_tag=new bool[file_no];

  for(i=0; i<file_no; ++i) {read_tag[i]=true;end_tag[i]=false; }

  int min=0;
  // begin to compare the record of the input files, select
  // the min one and put to the out file. if min==-1, that means
  // that no record in all the input file
  while(min!=-1) {
    min=-1;
    for(i=0; i<file_no; ++i) {
      if(read_tag[i]==true&&(!end_tag[i])) {
        tempI=infile[i]->getRecord(record[i],record_len[gb_step_no]);
		if(tempI==0) {
			end_tag[i]=true;
		}
		else {
	        read_tag[i]=false;
		}
      }
      if(read_tag[i]==false) {
        if(min==-1||tupleCmp(record[min],record[i])>0) 
          min=i;
      }
    }
    // generate the out file
    if(min!=-1) {
		if(end_flag) {
			temp=record[min];
			if(gb_step_no==0) {
/*
				fid=*((int*)(temp));
				eid=*((ushort*)(temp+sizeofInt));
				wfid=*((int*)(temp+sizeofInt+sizeofShort));
				weid=*((ushort*)(temp+2*sizeofInt+sizeofShort));
*/
				memcpy(&fid, temp, sizeofInt);
				memcpy(&eid, temp+sizeofInt, sizeofShort);
				memcpy(&wfid, temp+sizeofInt+sizeofShort, sizeofInt);
				memcpy(&weid, temp+2*sizeofInt+sizeofShort, sizeofShort);
				// write new file, get new yfid, yeid---------------+++++

 				char * ying_record;
                ushort ying_length; 

                DirPool::fileStore->getRecord(ying_record,ying_length, 
								wfid, weid);
				DirPool::fileStore->insertRecord(ying_record,ying_length, 
										wfid,weid);
				map->setValue(fid, eid, wfid, weid);
/*
printf("yuan: %d %d",DirPool::recordStore->getFid(ying_record, ying_length),
     				   DirPool::recordStore->getEid(ying_record, ying_length));					
printf("ying: %d %d   Old: %d, %d\n",wfid, weid, fid, eid);
*/

             	delete [] ying_record;
  
				memcpy(temp+sizeofInt+sizeofShort, (char*)(&wfid),sizeofInt);
				memcpy(temp+2*sizeofInt+sizeofShort, (char*)(&weid), 
									sizeofShort);
			}
			if(*old_fileNum==0) {
				*old_offset=btrees[gb_step_no]->addToEntry(temp);
				*old_fileNum=btrees[gb_step_no]->getFileNum();
				*record_count=1;
				memcpy(oldrecord, temp, record_len[gb_step_no]);
			}
			else if(tupleCmp1(oldrecord, temp)!=0) {
				btrees[gb_step_no]->addToLeaf(tempdata, oldrecord+gb_sort_offset);
				*old_offset=btrees[gb_step_no]->addToEntry(temp);
				*old_fileNum=btrees[gb_step_no]->getFileNum();
				*record_count=1;
				memcpy(oldrecord, temp, record_len[gb_step_no]);
			}
			else {
				btrees[gb_step_no]->addToEntry(temp);
				(*record_count)++;
			}
		}
		if(!end_flag)
	      	out->insertRecord(record[min],record_len[gb_step_no]);
      	read_tag[min]=true;
    }
  }
  if(end_flag) {
		btrees[gb_step_no]->addToLeaf(tempdata,oldrecord+gb_sort_offset);
		btrees[gb_step_no]->finish();
  }
		


  // delete the memory allocated
	delete [] oldrecord;
	delete [] tempdata;
	delete [] end_tag;
	delete [] read_tag;
	for(i=0; i<file_no; ++i) delete [] record[i];
	delete [] record;

  return;
}


///////////////////////////////////////////////////////////////////
// it is called by the method generateOutputFilename, it generate
// one file name used as the temperary filename according
// the pass no and the file no
///////////////////////////////////////////////////////////////////

void Sort::genFileName(int step_no,int pass_no, int file_no, 
					char*& filename, int tag) {
	char temp[30];
	if(tag==0) 
		sprintf(temp,"p_%d_%d_%d",step_no,pass_no,file_no);
	else
		sprintf(temp,"p__%d_%d_%d",step_no,pass_no,file_no);
	filename=new char[strlen(temp)+1];
	strcpy(filename,temp);
}


///////////////////////////////////////////////////////////////////
// This method is used to compare with the tuple in the Heapfile
// Two pointer to the tuple is passed in, return the -1, or 0 or 1
// to indicate which one is bigger, smaller, or equal to each other
// five variable can be used, 
// gb_sort_type, specify the type of the field that will be compared.
// gb_sort_order, specify the order of the result, Ascending or Descending
// gb_sort_offset, specify the offset of the field in the first tuple
// gb_sort_length, specify the length of the field that will be compared
///////////////////////////////////////////////////////////////////

int tupleCmp(const void *tuple1, const void* tuple2) {

  	int i;
  	char* temp1;
  	char* temp2;
	int tt1, tt2;
	int sizeofInt=sizeof(int);

  	temp1=(char*)((char*)tuple1+gb_sort_offset);
  	temp2=(char*)((char*)tuple2+gb_sort_offset);

	if(gb_sort_type==STRINGTYPE) {
		i=strncmp(temp1, temp2, gb_sort_size);
	}
	else { // INTTYPE

		memcpy(&tt1, temp1, sizeofInt);
		memcpy(&tt2, temp2, sizeofInt);

		i=tt1-tt2;
    // then the type must be attrInteger
	}

	if(i==0) {
		memcpy(&tt1, tuple1, sizeofInt);
		memcpy(&tt2, tuple2, sizeofInt);
		i=tt1-tt2;
	}
	if(i==0) {
		ushort t1, t2;
		int sizeofShort=sizeof(short);

		memcpy(&t1, (char*)(tuple1)+sizeofInt, sizeofShort);
		memcpy(&t2, (char*)(tuple2)+sizeofInt, sizeofShort);
		i=t1-t2;
	}

	if(gb_sort_order==ASCENDING) {
		if(i<0) return -1;
		else if(i==0) return 0;
		else return 1;
	}
	else {
		// descending
		if(i<0) return 1;
		else if(i==0) return 0;
		else return -1;
	}
}

int tupleCmp1(const void *tuple1, const void* tuple2) {

  	int i;
  	char* temp1;
  	char* temp2;
	int tt1, tt2;
	int sizeofInt=sizeof(int);

  	temp1=(char*)((char*)tuple1+gb_sort_offset);
  	temp2=(char*)((char*)tuple2+gb_sort_offset);

	if(gb_sort_type==STRINGTYPE) {
		i=strncmp(temp1, temp2, gb_sort_size);
	}
	else { // INTTYPE

		memcpy(&tt1, temp1, sizeofInt);
		memcpy(&tt2, temp2, sizeofInt);

		i=tt1-tt2;
    // then the type must be attrInteger
	}

	if(gb_sort_order==ASCENDING) {
		if(i<0) return -1;
		else if(i==0) return 0;
		else return 1;
	}
	else {
		// descending
		if(i<0) return 1;
		else if(i==0) return 0;
		else return -1;
	}
}
